#ifndef SCUMMVM_H
#define SCUMMVM_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sstream>

#define WRITE_OUTPUT

#define BASE_DIR	"."
#define INPUT_DIR	BASE_DIR "/sounds"
#define OUTPUT_DIR	BASE_DIR "/sounds/convert"

const double SAMPLE_RATE = 44100.0;

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned int uint;

const int rtSound = 0;

/*********************************
 * WAVE-file data structs (START)
 *********************************/

#pragma pack(1)

typedef byte riffChunkID_t[4];

typedef struct riffChunk_t {
    riffChunkID_t id;
    uint dataSize;
} riffChunk_t;

typedef struct riffHeader_t {
    riffChunk_t chunkInfo;
    riffChunkID_t riffType;
} riffHeader_t;

typedef struct waveFmtChunk_t {
    riffChunk_t chunkInfo;
    word formatTag;
    word numChannels;
    uint samplesPerSec;
    uint avgBytesPerSec;
    word blockAlign;
    word bitsPerSample;
} waveFmtChunk_t;

#pragma pack()

/*********************************
 * WAVE-file data structs (END)
 *********************************/

class Player_AppleII {
public:
	Player_AppleII() : m_speakerState(0) {}
	
	void startSound(int nr /*x*/);

private:
	FILE *m_out;
	byte m_speakerState;
	size_t m_dataSize;

private:
	void speakerToggle();
	void generateSamples(int cycles);
	void wait(int interval, int count);
	byte noise();

	void soundFunc1(const byte *params);
	void _soundFunc1(int interval, int count);
	void soundFunc2(const byte *params);
	void _soundFunc2(int interval, int count);
	void soundFunc3(const byte *params);
	void _soundFunc3(int interval, int count);
	void soundFunc4(const byte *params);
	void _soundFunc4(byte param0, byte param1, byte param2);
	void soundFunc5(const byte *params);
	void _soundFunc5(int interval, int count);

	void prepareOutput(int nr) {
		#ifdef WRITE_OUTPUT

		m_dataSize = 0;

		std::stringstream s;
		s << OUTPUT_DIR << "/test" << nr << ".wav";
		m_out = fopen(s.str().c_str(), "wb");

		byte headerDummy[sizeof(riffHeader_t) + sizeof(waveFmtChunk_t) + sizeof(riffChunk_t)];
		fwrite(headerDummy, sizeof(headerDummy), 1, m_out);

		#endif
	}

	void finishOutput() {
		#ifdef WRITE_OUTPUT

		const int channels = 1;
		const int WAVE_FORMAT_PCM = 1; // PCM (uncompressed)

		fseek(m_out, 0, SEEK_SET);

		riffHeader_t waveHdr;
		const riffChunkID_t RIFF_CHUNK_HDR = { 'R', 'I', 'F', 'F' };
		memcpy(waveHdr.chunkInfo.id, RIFF_CHUNK_HDR, sizeof(RIFF_CHUNK_HDR));
		waveHdr.chunkInfo.dataSize = (sizeof(riffHeader_t) - 8) +
			sizeof(waveFmtChunk_t) + sizeof(riffChunk_t) + m_dataSize;
		const riffChunkID_t RIFF_TYPE_WAVE = { 'W', 'A', 'V', 'E' };
		memcpy(waveHdr.riffType, RIFF_TYPE_WAVE, sizeof(RIFF_TYPE_WAVE));
		fwrite(&waveHdr, sizeof(waveHdr), 1, m_out);

		waveFmtChunk_t waveFmt;
		const riffChunkID_t RIFF_CHUNK_FMT = { 'f', 'm', 't', ' ' };
		memcpy(waveFmt.chunkInfo.id, RIFF_CHUNK_FMT, sizeof(RIFF_CHUNK_FMT));
		waveFmt.chunkInfo.dataSize = sizeof(waveFmtChunk_t) - 8;
		waveFmt.formatTag = WAVE_FORMAT_PCM;
		waveFmt.numChannels = channels;
		waveFmt.samplesPerSec = (int)SAMPLE_RATE;
		waveFmt.avgBytesPerSec = (int)SAMPLE_RATE * channels * 1;
		waveFmt.blockAlign = channels * 1;
		waveFmt.bitsPerSample = channels * 8;
		fwrite(&waveFmt, sizeof(waveFmt), 1, m_out);

		riffChunk_t dataChunk;
		const riffChunkID_t RIFF_CHUNK_DATA = { 'd', 'a', 't', 'a' };
		memcpy(dataChunk.id, RIFF_CHUNK_DATA, sizeof(RIFF_CHUNK_DATA));
		dataChunk.dataSize = m_dataSize;
		fwrite(&dataChunk, sizeof(dataChunk), 1, m_out);

		fclose(m_out);
		#endif
	}
};

class VirtualMachine {
private:
	byte *m_data;
public:
	VirtualMachine() : m_data(NULL) {}

	byte* getResourceAddress(int rtSound, int nr) {
		free(m_data);

		std::stringstream s;
		s << INPUT_DIR << "/sound" << nr << ".snd";
		FILE* f = fopen(s.str().c_str(), "rb");
		if (!f) {
			return NULL;
		}
		byte sizeBytes[2];
		if (fread(sizeBytes, 1, 2, f) != 2) {
			fclose(f);
			return NULL;
		}
		word size = (sizeBytes[1] << 8) | sizeBytes[0];
		m_data = (byte*)malloc(size);
		fseek(f, 0, SEEK_SET);
		if (fread(m_data, 1, size, f) != size) {
			fclose(f);
			return NULL;
		}
		fclose(f);

		return m_data;
	}
};

extern VirtualMachine *_vm;

#endif // SCUMMVM_H